import os
import asyncio
import urllib.parse as url_parser

from .model import OVInfo, CmdResponse


class VPNManager:

    def __init__(self, url: str):
        self._host = None
        self._port = None
        self._unix_socket = None
        self._timeout_read = 1  # second

        self._stream_reader: asyncio.StreamReader = None
        self._stream_writer: asyncio.StreamWriter = None

        self._parse_url(url)
        self._check_connect()

    def _parse_url(self, url: str):
        """Parses the url and fills in the class fields

        Raises:
            ValueError
        """
        obj = url_parser.urlparse(url)
        if obj.scheme == "unix":
            if obj.netloc:
                raise ValueError(
                    "Wrong format for unix socket, need: unix:///...")
            self._unix_socket = obj.path
        elif obj.scheme == "tcp":
            self._host = obj.hostname
            self._port = obj.port
        else:
            raise ValueError(f"Schema not found: {obj.scheme}")

    def _check_connect(self):
        if (
            (self._host is None or self._port is None) and
            self._unix_socket is None
        ):
            raise ValueError("Host or unix_socket must be set")

    async def send_command(self, command: str) -> str:
        """Send command to openvpn management interface and return response"""
        if self._stream_reader is None or self._stream_writer is None:
            await self.connect()

        self._stream_writer.write(command.encode())
        await self._stream_writer.drain()
        return await self._read()

    async def connect(self):
        if self._unix_socket is not None:
            reader, writer = await asyncio.open_unix_connection(
                self._unix_socket
            )
        else:
            reader, writer = await asyncio.open_connection(
                self._host, self._port
            )
        self._stream_reader, self._stream_writer = reader, writer

        # read hello
        await self._read()

    async def _read(self):
        result = ""
        try:
            while True:
                a: bytes = await asyncio.wait_for(
                    self._stream_reader.read(1024), timeout=self._timeout_read
                )
                result += a.decode()
        except asyncio.exceptions.TimeoutError:
            pass
        return result

    async def close(self):
        self._stream_writer.close()
        await self._stream_writer.wait_closed()

    async def cmd_status(self) -> OVInfo:
        """Send status command to openvpn management interface

        Format, --status-version [n], we uses 2
        2 -- a more reliable format for external processing.
        Compared to version 1, the client list contains some additional fields:
        Virtual Address, Virtual IPv6 Address, Username, Client ID, Peer ID.
        Future versions may extend the number of fields.
        """
        return OVInfo.create_from_str(await self.send_command("status 2\n"))

    async def cmd_kill(self, common_name: str) -> CmdResponse:
        """Send kill command to openvpn management interface

        Example:
        SUCCESS: common name 'd.test' found, 1 client(s) killed

        """
        answer = await self.send_command(f"kill {common_name}\n")

        answer = answer.strip()
        if answer.startswith("SUCCESS:"):
            return CmdResponse(status="SUCCESS", message=answer[8:].strip())
        elif answer.startswith("ERROR:"):
            return CmdResponse(status="ERROR", message=answer[6:].strip())
        return CmdResponse(status="ERROR", message=f"Unknown error: {answer}")



import dataclasses
from enum import Enum
from datetime import datetime
from typing import List, Optional


class Status(Enum):
    """Status of openvpn management interface"""
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


@dataclasses.dataclass
class CmdResponse:
    """Answer from openvpn management interface"""
    status: Status
    message: str


@dataclasses.dataclass
class Client:
    """Openvpn client"""
    common_name: str
    real_address: str
    virtual_address: str
    virtual_ipv6_address: str
    bytes_received: int
    bytes_sent: int
    connected_since: datetime
    username: str
    client_id: int
    peer_id: int


@dataclasses.dataclass
class OVInfo:
    """Openvpn info"""
    title: str
    time: Optional[datetime] = None
    clients: List[Client] = dataclasses.field(default_factory=list)

    @staticmethod
    def create_from_str(data: str) -> "OVInfo":
        """Parse openvpn info

        Raises:
            ValueError: If error in format

        Example:
        TITLE,OpenVPN 2.4.12 x86_64-redhat-linux-gnu [Fedora EPEL patched] [SSL (OpenSSL)] [LZO] [LZ4] [EPOLL] [PKCS11] [MH/PKTINFO] [AEAD] built on Mar 17 2022
        TIME,Fri Mar 31 15:54:52 2023,1680267292
        HEADER,CLIENT_LIST,Common Name,Real Address,Virtual Address,Virtual IPv6 Address,Bytes Received,Bytes Sent,Connected Since,Connected Since (time_t),Username,Client ID,Peer ID
        CLIENT_LIST,d.test,25.13.19.12:1194,172.20.100.17,,18229,7921,Fri Mar 31 15:54:33 2023,1680267273,UNDEF,7310,0
        HEADER,ROUTING_TABLE,Virtual Address,Common Name,Real Address,Last Ref,Last Ref (time_t)
        ROUTING_TABLE,172.20.100.17,d.test,95.53.79.72:1194,Fri Mar 31 15:54:44 2023,1680267284
        GLOBAL_STATS,Max bcast/mcast queue length,0
        END
        """
        title = ""
        time = None
        clients = []

        for line in data.splitlines():
            line = line.strip()

            if line.startswith("TITLE,"):
                title = line.split(",", 2)[1]
                continue
            if line.startswith("TIME,"):
                time = datetime.strptime(
                    line.split(",", 2)[1], "%a %b %d %H:%M:%S %Y")
                continue
            if line.startswith("CLIENT_LIST,"):
                line = line.split(",")
                client = Client(
                    common_name=line[1],
                    real_address=line[2].split(":")[0],  # 1.2.3.4:1194
                    virtual_address=line[3],
                    virtual_ipv6_address=line[4],
                    bytes_received=int(line[5]),
                    bytes_sent=int(line[6]),
                    connected_since=datetime.strptime(
                        line[7], "%a %b %d %H:%M:%S %Y"
                    ),
                    username=line[9],
                    client_id=int(line[10]),
                    peer_id=int(line[11]),
                )
                clients.append(client)
                continue
        return OVInfo(title=title, time=time, clients=clients)
